#ifdef PETSC_RCS_HEADER
static char vcid[] = "$Id: matreg.c,v 1.1 1999/05/08 16:53:40 bsmith Exp bsmith $";
#endif
/*
     Mechanism for register PETSc matrix types
*/
#include "src/sles/mat/matimpl.h"      /*I "mat.h" I*/
#include "sys.h"

int MatRegisterAllCalled = 0;

/*
   Contains the list of registered Mat routines
*/
FList MatList = 0;

#undef __FUNC__  
#define __FUNC__ "MatSetType"
/*@C
   MatSetType - Builds matrix object for a particular matrix type

   Collective on Mat

   Input Parameters:
+  mat      - the matrix object
-  matype   - matrix type

   Options Database Key:
.  -mat_type  <method> - Sets the type; use -help for a list 
    of available methods (for instance, seqaij)

   Notes:  
   See "${PETSC_DIR}/include/mat.h" for available methods

  Level: intermediate

.keywords: Mat, set, method

.seealso: PCSetType(), VecSetType()
@*/
int MatSetType(Mat mat,MatType matype)
{
  int ierr,(*r)(Mat);

  PetscFunctionBegin;
  PetscValidHeaderSpecific(mat,MAT_COOKIE);

  if (PetscTypeCompare(mat->type_name,matype)) PetscFunctionReturn(0);

  if (mat->setupcalled) {
    /* destroy the old private Mat context */
    ierr = (*mat->ops->destroy)(Mat);CHKERRQ(ierr);
    mat->data = 0;
  }
  /* Get the function pointers for the iterative method requested */
  if (!MatRegisterAllCalled) {ierr = MatRegisterAll(PETSC_NULL);CHKERRQ(ierr);}

  ierr =  FListFind(mat->comm, MatList, matype,(int (**)(void *)) &r );CHKERRQ(ierr);

  if (!r) SETERRQ1(1,1,"Unknown Mat type given: %s",matype);

  mat->data        = 0;
  mat->setupcalled = 0;
  ierr = (*r)(mat);CHKERRQ(ierr);

  if (mat->type_name) PetscFree(mat->type_name);
  mat->type_name = (char *) PetscMalloc((PetscStrlen(matype)+1)*sizeof(char));CHKPTRQ(mat->type_name);
  ierr = PetscStrcpy(mat->type_name,matype);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNC__  
#define __FUNC__ "MatRegisterDestroy"
/*@C
   MatRegisterDestroy - Frees the list of matrix types that were
   registered by MatRegister().

   Not Collective

   Level: advanced

.keywords: Mat, register, destroy

.seealso: MatRegister(), MatRegisterAll()
@*/
int MatRegisterDestroy(void)
{
  int ierr;

  PetscFunctionBegin;
  if (MatList) {
    ierr = FListDestroy( MatList );CHKERRQ(ierr);
    MatList = 0;
  }
  MatRegisterAllCalled = 0;
  PetscFunctionReturn(0);
}

#undef __FUNC__  
#define __FUNC__ "MatGetType"
/*@C
   MatGetType - Gets the matrx type as a string from the matrix object.

   Not Collective

   Input Parameter:
.  mat - the matrix

   Output Parameter:
.  name - name of matrix type

   Level: intermediate

.keywords: Mat, get, method, name

.seealso: MatSetType()
@*/
int MatGetType(Mat mat,MatType *type)
{
  PetscFunctionBegin;
  *type = mat->type_name;
  PetscFunctionReturn(0);
}

#undef __FUNC__  
#define __FUNC__ "MatSetTypeFromOptions"
/*@
   MatSetTypeFromOptions - Sets Mat type from the options database, if not
       given then sets default.

   Collective on Mat

   Input Parameters:
.  Mat - the Krylov space context

   Level: developer

.keywords: Mat, set, from, options, database

.seealso: MatPrintHelp(), MatSetFromOptions(), SLESSetFromOptions(),
          SLESSetTypeFromOptions()
@*/
int MatSetTypeFromOptions(Mat Mat)
{
  int       flg, ierr;
  char      method[256];

  PetscFunctionBegin;
  PetscValidHeaderSpecific(Mat,Mat_COOKIE);

  ierr = OptionsGetString(Mat->prefix,"-Mat_type",method,256,&flg);
  if (flg) {
    ierr = MatSetType(Mat,method);CHKERRQ(ierr);
  }
  /*
    Set the type if it was never set.
  */
  if (!Mat->type_name) {
    ierr = MatSetType(Mat,MatGMRES);CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

/*MC
   MatRegister - Adds a method to the Krylov subspace solver package.

   Synopsis:
   MatRegister(char *name_solver,char *path,char *name_create,int (*routine_create)(Mat))

   Not Collective

   Input Parameters:
+  name_solver - name of a new user-defined solver
.  path - path (either absolute or relative) the library containing this solver
.  name_create - name of routine to create method context
-  routine_create - routine to create method context

   Notes:
   MatRegister() may be called multiple times to add several user-defined solvers.

   If dynamic libraries are used, then the fourth input argument (routine_create)
   is ignored.

   Sample usage:
.vb
   MatRegister("my_solver",/home/username/my_lib/lib/libO/solaris/mylib.a,
               "MySolverCreate",MySolverCreate);
.ve

   Then, your solver can be chosen with the procedural interface via
$     MatSetType(Mat,"my_solver")
   or at runtime via the option
$     -Mat_type my_solver

   Level: advanced

   $PETSC_ARCH and $BOPT occuring in pathname will be replaced with appropriate values.

.keywords: Mat, register

.seealso: MatRegisterAll(), MatRegisterDestroy()

M*/

#undef __FUNC__  
#define __FUNC__ "MatRegister_Private"
int MatRegister_Private(char *sname,char *path,char *name,int (*function)(Mat))
{
  int  ierr;
  char fullname[256];

  PetscFunctionBegin;
  ierr = PetscStrcpy(fullname,path);CHKERRQ(ierr);
  PetscStrcat(fullname,":");PetscStrcat(fullname,name);
  ierr = FListAdd_Private(&MatList,sname,fullname,(int (*)(void*))function);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
