/*$Id: ex18.c,v 1.4 2000/01/11 21:02:20 bsmith Exp $*/
/*
Laplacian in 3D. Modeled by the partial differential equation

   Laplacian u = 0, 0 < x,y,z < 1,

with boundary conditions

   u = 1 for x = 0, x = 1, y = 0, y = 1, z = 0, z = 1.

    Contributed by Kashif Rasul of Steve's great home country
*/

static char help[] = "Solves Laplacian in 3D with a user defined\n\
Newton method that uses SLES. The command line options are:\n\
   -mx <xg>, where <xg> = number of grid points in the x-direction\n\
   -my <yg>, where <yg> = number of grid points in the y-direction\n\
   -mz <zg>, where <zg> = number of grid points in the z-direction\n\n";

#include "da.h"
#include "sles.h"

typedef struct{
  DA  da;
} AppCtx;

extern int ComputeJacobian(AppCtx*,Mat);

int main(int argc,char **argv)
{
  MPI_Comm     comm;
  SLES         sles;
  Vec          X,Y,F;
  Mat          J;
  AppCtx       user;
  int          Nx = PETSC_DECIDE,Ny = PETSC_DECIDE,Nz = PETSC_DECIDE;
  int          ierr,N;
  int          its,m,mx,my,mz;
  double       norm;
  Scalar       one = 1.0,mone = -1.0;
  PetscTruth   no_output;

  PetscInitialize(&argc,&argv,(char *)0,help);
  comm = PETSC_COMM_WORLD;

  ierr = OptionsHasName(PETSC_NULL,"-no_output",&no_output);CHKERRA(ierr);
  
  mx = 4; my = 4; mz = 4; N = mx*my*mz;
  ierr = OptionsGetInt(PETSC_NULL,"-mx",&mx,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-my",&my,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-mz",&mz,PETSC_NULL);CHKERRA(ierr);
  
  ierr = OptionsGetInt(PETSC_NULL,"-Nx",&Nx,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-Ny",&Ny,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-Nz",&Nz,PETSC_NULL);CHKERRA(ierr);
  ierr = DACreate3d(comm,DA_NONPERIODIC,DA_STENCIL_BOX,mx,my,mz,Nx,Ny,Nz,1,1,PETSC_NULL,PETSC_NULL,PETSC_NULL,&user.da);CHKERRA(ierr);
  ierr = DACreateGlobalVector(user.da,&X);CHKERRA(ierr);
  ierr = VecDuplicate(X,&F);CHKERRA(ierr);
  ierr = VecDuplicate(X,&Y);CHKERRA(ierr);
  ierr = VecGetLocalSize(X,&m);CHKERRA(ierr);
  ierr = MatCreateMPIAIJ(comm,m,m,N,N,7,PETSC_NULL,5,PETSC_NULL,&J);

  /* set right hand side of all one */
  ierr = VecSet(&one,F);CHKERRA(ierr);
  ierr = SLESCreate(comm,&sles);CHKERRA(ierr);
  ierr = SLESSetFromOptions(sles);CHKERRA(ierr);
  
  ierr = ComputeJacobian(&user,J);CHKERRA(ierr);
  ierr = SLESSetOperators(sles,J,J,DIFFERENT_NONZERO_PATTERN);CHKERRA(ierr);
  ierr = SLESSolve(sles,F,X,&its);CHKERRA(ierr);

  /* compute residual */
  ierr = MatMult(J,X,Y);CHKERRA(ierr);
  ierr = VecAXPY(&mone,F,Y);CHKERRA(ierr);
  ierr = VecNorm(Y,NORM_2,&norm);CHKERRA(ierr);

  if (!no_output) {
    ierr = PetscPrintf(comm,"Residual norm %A iterations = %d,\n",norm,its);CHKERRA(ierr);
  }

  ierr = MatDestroy(J);CHKERRA(ierr);
  ierr = VecDestroy(Y);CHKERRA(ierr);
  ierr = VecDestroy(X);CHKERRA(ierr);
  ierr = VecDestroy(F);CHKERRA(ierr);
  ierr = SLESDestroy(sles);CHKERRA(ierr);
  ierr = DADestroy(user.da);CHKERRA(ierr);
  
  PetscFinalize();

  return 0;
}

    
int ComputeJacobian(AppCtx *user,Mat jac)
{
  int    *ltog,ierr,i,j,k,mx,my,mz,xm,ym,zm,xs,ys,zs,Xm,Ym,Zm,Xs,Ys,Zs,row,nloc,col[7],base1,grow;
  Scalar two = 2.0,one = 1.0,v[7],Hx,Hy,Hz,HxHydHz,HyHzdHx,HxHzdHy;

  ierr = DAGetInfo(user->da,0,&mx,&my,&mz,0,0,0,0,0,0,0);CHKERRQ(ierr);  
  Hx = one / (double)(mx-1); Hy = one / (double)(my-1); Hz = one / (double)(mz-1);
  HxHydHz = Hx*Hy/Hz; HxHzdHy = Hx*Hz/Hy; HyHzdHx = Hy*Hz/Hx;
  ierr = DAGetCorners(user->da,&xs,&ys,&zs,&xm,&ym,&zm);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&Xs,&Ys,&Zs,&Xm,&Ym,&Zm);CHKERRQ(ierr);
  ierr = DAGetGlobalIndices(user->da,&nloc,&ltog);CHKERRQ(ierr);
  
  for (k=zs; k<zs+zm; k++){
    base1 = (k-Zs)*(Xm*Ym);
    for (j=ys; j<ys+ym; j++){
      row = base1 + (j-Ys)*Xm + xs - Xs - 1;
      for(i=xs; i<xs+xm; i++){
	row++;
	grow = ltog[row];
	if (i==0 || j==0 || k==0 || i==mx-1 || j==my-1 || k==mz-1){
	  ierr = MatSetValues(jac,1,&grow,1,&grow,&one,INSERT_VALUES);   CHKERRQ(ierr);
	  continue;
	}
	v[0] = -HxHydHz; col[0] = ltog[row - Xm*Ym];
	v[1] = -HxHzdHy; col[1] = ltog[row - Xm];
	v[2] = -HyHzdHx; col[2] = ltog[row - 1];
	v[3] = two*(HxHydHz + HxHzdHy + HyHzdHx); col[3]=grow;
	v[4] = -HyHzdHx; col[4] = ltog[row + 1];
	v[5] = -HxHzdHy; col[5] = ltog[row + Xm];
	v[6] = -HxHydHz; col[6] = ltog[row + Xm*Ym];
	ierr = MatSetValues(jac,1,&grow,7,col,v,INSERT_VALUES);CHKERRQ(ierr);
      }
    }
  }
  ierr = MatAssemblyBegin(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  return 0;
}
